/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#include "flag.h"
#include "oi_wavelength.h"
#include "oi_vis.h"
#include "oi_target.h"
#define clight  2.9979250e8 /*m per sec */
#define twopi   6.2831853072
#define NBIN 16
#define WAVELENGTHFIX 0.990574

/*
    Usage::
    oirAverageVis infile flagFile outFile 
*/
void Usage() {
   printf("   oirAverageVis inFile flagFile outFile\n");
   printf("   inFile = file name of output of oirRotateGroupDelay\n");
   printf("   flagFile = file name of output of oirAutoFlag\n");
   printf("   outFile = output file name containing averaged correlated flux\n");
   return;
}

void getUVW(fitsfile *inFile, float *u, float *v, float *w, int *ierr) {

   float x1, x2, y1, y2, z1, z2;
   float dx, dy, dz;
   float ra, dec, lst, ha;
   float vltlat;
   float radeg = 57.29577951;
   float cha, sha, cdec, sdec, clat, slat;
   int i;
   int iKey;

   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO ISS CONF T1X", &x1, 0, ierr);
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO ISS CONF T2X", &x2, 0, ierr);
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO ISS CONF T1Y", &y1, 0, ierr);
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO ISS CONF T2Y", &y2, 0, ierr);
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO ISS CONF T1Z", &z1, 0, ierr);
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO ISS CONF T2Z", &z2, 0, ierr);
   i = fits_read_key(inFile, TFLOAT, "RA", &ra, 0, ierr);
   i = fits_read_key(inFile, TFLOAT, "DEC", &dec, 0, ierr);
   i = fits_read_key(inFile, TFLOAT, "LST", &lst, 0, ierr);
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO ISS GEOLAT", &vltlat, 0, ierr);
   ra  = ra/radeg;
   dec = dec/radeg;
   lst = lst*15./(3600.*radeg); /* LST is in sec of time */
   ha = lst - ra;
   vltlat = vltlat/radeg;
   dx = x1 - x2;
   dy = y1 - y2;
   dz = z1 - z2;

   cha = cos(ha);
   sha = sin(ha);
   cdec = cos(dec);
   sdec = sin(dec);
   clat = cos(vltlat);
   slat = sin(vltlat);

   *u = dx*cha - dy * slat * sha + dz * clat*sha;
   *v = dx * sha *sdec + dy *( slat*cha*sdec + clat*cdec) - dz*(clat*cha*sdec - slat*cdec);
   *w = -dx*sha*cdec - dy*(slat*cha*cdec - clat*sdec) + dz*(clat*cha*cdec+slat*sdec);
   return;
} /* end of getUVW */

int main (int argc, char** argv) {

   int i, j, ierr;
   long nRow, iRow ;
   int nx, nWave ;
   float dWave;
   float *frequency, *wave, *band;
   float u, v, w;

   char outFileName[130];
   char *bang="!";

   fitsfile *inFile;
   fitsfile *flagFile;
   fitsfile *outFile = NULL;
   imaging_fdata_table_def *idata;
   imaging_detector_table_def *outDetectorDef;
   imaging_fdata_row *inData;

   wavelength_table_def *wTable;
   wavelength_table_row *wRow;

   flag_table_def *fTable;
   flag_table_row *fRow;

   vis_table_def *vTable;
   vis_table_row *vRow;

   target_table_def *targTable;
   target_table_row *targRow;

   int   iFlag;
   long  nFlag;
   double* flagTimes;
   int   *goodData;
   float *binData;
   int   nBin, iBin, goodBin;
   int   doErrors;
   float ss, cc, cs, c0, s0, a0;

   double *aveVis; /* complex average Vis */
   double *visAmp; /* amplitude average Vis */
   double *visPhase; /* phase average Vis(deg) */
   double *errAmp;   /* rms amplitude err */
   double *errPhase; /* rms phase error(deg) */
   double degra;

   float  equinox;
   double ra, dec;
   char   targName[80];

   int    nGood, good, flagGood;
   float  totalExposure;
   double visTime;

   ierr = 0;
/*
      open input file 
*/
   if (argc < 3) {
      Usage();
      return 1;
   }
   printf("      ****** oirAverageVis ******\n");

/*
      read in NGOOD from flag table
*/
   fits_open_file (&flagFile, argv[2], READONLY, &ierr);
   if (ierr != 0) {
      printf("fits openfile of flagfile %s gave error %i\n",argv[2],
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
   i = fits_read_key(flagFile, TINT, "NGOOD", &flagGood, 0, &ierr);
   i = fits_close_file(flagFile, &ierr);
/*
 *     now open ungroupdelay data file
*/
   fits_open_file (&inFile, argv[1], READONLY, &ierr);
   if (ierr != 0) {
      printf("fits openfile of %s gave error %i\n",argv[1],
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   } else 
      printf("   Reading ungroupdelay data from %s\n", argv[1]);
/*
   get uvw
*/
   getUVW(inFile, &u, &v, &w, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "      error reading uvw from %s %i\n",argv[1], ierr);
      fprintf (stderr, "      setting uvw to 0.\n");
      u = 0.;
      v = 0.;
      w = 0.;
      ierr = 0;
   }
/*
 *    get ra, dec, epoch, target name
*/
   i = fits_read_key(inFile, TFLOAT, "EQUINOX", &equinox, 0, &ierr);
   i = fits_read_key(inFile, TDOUBLE, "RA", &ra, 0, &ierr);
   i = fits_read_key(inFile, TDOUBLE, "DEC", &dec, 0, &ierr);
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO OBS TARG NAME", 
      targName, 0, &ierr);
/*
      create output file and copy header
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[3]);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n",argv[3], ierr);
      printf("%s \n",strerror(errno));
      return 1;
   } else
      printf("   Writing averaged data into     %s\n", argv[3]);
   i = fits_copy_header (inFile, outFile, &ierr);
   i = fits_write_key(outFile, TINT, "NGOOD", (void*)(&flagGood),
      "#FRAMES FLAGGED OK", &ierr);
/*
      get detector info
*/

   outDetectorDef = read_imaging_detector_table(inFile, &ierr);
   idata          = open_imaging_fdata_table(inFile, &ierr);
   inData         = make_imaging_fdata_row(idata);

/*     
   get row size (pseudo-complex)
*/
   nx = outDetectorDef->naxis[0][0];
   nWave = nx/2; 
   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
/*       
      get frequency information from detector table
*/
   frequency = imaging_detector_wcs(outDetectorDef, 1);
   wave = (float *)malloc(nWave * sizeof(float));
   band = (float *)malloc(nWave * sizeof(float));
   for (i=0; i<nWave; i++) wave[i] = WAVELENGTHFIX * clight/frequency[i];
   printf("   Applying correction of %f to input wavelength scale.\n", WAVELENGTHFIX);
   /* cludge */
   dWave = (wave[nWave-1] - wave[0])/nWave;
   if (dWave < 0.) dWave = - dWave;
   for (i=0; i<nWave; i++) band[i] = dWave;

/*
     set up wavelength table in output file
*/
   wTable = make_wavelength_table_def(outFile,
      nWave, "MIDI/PRISM", &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error creating wavelength table %i\n", ierr);
      return 1;
   }
   wRow = make_wavelength_table_row(wTable);

   for (iRow=0; iRow<nWave; iRow++) {
      wRow->wave = wave[iRow];
      wRow->band = band[iRow];
      i = write_wavelength_table(outFile, wTable, wRow, iRow+1, &ierr);
   }
   kill_wavelength_table_row(wRow);
   kill_wavelength_table_def(wTable);
/*
 *       set up target table
*/ 
   targTable = make_target_table_def(outFile, &ierr);
   targRow   = make_target_table_row(targTable);
   targRow->target_id = 1;
   targRow->raep0     = ra;
   targRow->decep0    = dec;
   targRow->equinox   = equinox;
   strncpy(targRow->target, targName, 16);
   i = write_target_table(outFile, targTable, targRow, 1, &ierr);
   kill_target_table_row(targRow);
   kill_target_table_def(targTable);
   
/*
      read in flag data and expand
*/
   fits_open_file (&flagFile, argv[2], READONLY, &ierr);
   if (ierr != 0) {
      printf("fits openfile of flagfile %s gave error %i\n",argv[2],
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   } else
      printf("   Getting flag data from         %s\n", argv[2]);
   fTable  = open_flag_table(flagFile, &ierr);
   if (ierr != 0) {
      printf("create flag table gave error %i\n", ierr);
      return 1;
   }
   fRow = make_flag_table_row(fTable);
   i = fits_get_num_rows (flagFile,  &nFlag, &ierr); 
   flagTimes = (double *)malloc(nFlag * 2 * sizeof(double));
   for (iRow =0; iRow <nFlag; iRow++) {
      i=read_flag_table(flagFile, fTable, fRow, iRow+1, &ierr);
      flagTimes[2*iRow]   = fRow->timeRange[0];
      flagTimes[2*iRow+1] = fRow->timeRange[1];
   }
   i = close_flag_table(flagFile, fTable, &ierr);
   kill_flag_table_row(fRow);
/*    
      allocate memory to store statistical variables
*/
   aveVis   = (double *)malloc(nx * sizeof(double));
   visAmp   = (double *)malloc(nWave * sizeof(double));
   visPhase = (double *)malloc(nWave * sizeof(double));
   errAmp   = (double *)malloc(nWave * sizeof(double));
   errPhase = (double *)malloc(nWave * sizeof(double));

   memset(aveVis, 0, nx*sizeof(double));

   nGood         = 0;
   totalExposure = 0.;
   goodData = (int *)malloc(nRow * sizeof(int));
/*
         read through data once and set up flag table
         check against flag times; it would be faster if we sorted 
         times first; average visibilities coherently
*/
   for (iRow = 1; iRow <= nRow; iRow++) {
      i = read_imaging_fdata_table(inFile, idata, inData, iRow, &ierr);
      good    = 1;
      visTime = inData->time;
      if (isnan(visTime)) good = 0; else 
         for (iFlag =0; iFlag<nFlag && (good == 1); iFlag++) 
            if (visTime >= flagTimes[2*iFlag] && (visTime <= flagTimes[2*iFlag+1])) good=0;
      goodData[iRow-1] = good;
      if (good == 1) {
         nGood = nGood + 1;
         totalExposure = totalExposure + inData->exp;
         for (i=0;i<nx;i++) aveVis[i] = aveVis[i] + inData->data[0][i];
      }
   }
   if (nGood > 0) {
      for (i=0;i<nx;i++) aveVis[i] = aveVis[i]/nGood;
      printf("   Averaged %i so-called good frames.\n", nGood);
   }  else {
         printf("   No unflagged data in input.\n");
         return 1;
   }
/*
         now estimate the noise:
         create "sample buffer" to bin input data
         read the data again, averaging into bins
         these we will use to estimate lower frequency
         noise (high freq noise is mostly due to side band problems
*/
   nBin    = NBIN;
   binData = (float *)malloc(nBin * nx * sizeof(float));
   goodBin = (int)(nGood/nBin);
   doErrors = (goodBin >= 4);
   if (doErrors) { /* want at least 4 good frames per bin for error estimates */
      nGood   = 0;
      for (iRow = 1; iRow <= nRow; iRow++) {
         if (goodData[iRow-1]) {
            i = read_imaging_fdata_table(inFile, idata, inData, iRow, &ierr);
            iBin = (int)(nGood++/goodBin);
            if (iBin < nBin) 
               for (i=0;i<nx;i++) binData[nx*iBin +i] = binData[nx*iBin+i] + inData->data[0][i];
         } /* loop over rows */
      } /* this row OK */
   }  /* enough frames to compute errors */
/*
          calculate average amplitude and phase
*/
   degra = 360/twopi;
   for (i=0;i<nWave;i++) {
      visAmp[i]   = sqrt(aveVis[2*i]*aveVis[2*i] + aveVis[2*i+1]*aveVis[2*i+1]);
      visPhase[i] = degra * atan2(aveVis[2*i+1], aveVis[2*i]);
   }
/*
          now average binned data, subtract global average,
          take 2nd moments of cosine/sine residuals
          and use averages of these to estimate errors in 
          both amplitude and phase
*/
   if (doErrors) {
      for (i=0;i<nWave;i++) {
         cc = 0.; /* cosine^2 residual */
         ss = 0.; /* sine^2 */
         cs = 0.; /* cosine * sine */
         for (iBin = 0; iBin<nBin; iBin++) {
            binData[nx*iBin+2*i]   = binData[nx*iBin+2*i]/(float)goodBin;
            binData[nx*iBin+2*i+1] = binData[nx*iBin+2*i+1]/(float)goodBin;
            binData[nx*iBin+2*i]   = binData[nx*iBin+2*i]   - aveVis[2*i];
            binData[nx*iBin+2*i+1] = binData[nx*iBin+2*i+1] - aveVis[2*i+1];
            cc = cc + binData[nx*iBin+2*i]   * binData[nx*iBin+2*i];
            ss = ss + binData[nx*iBin+2*i+1] * binData[nx*iBin+2*i+1];
            cs = cs + binData[nx*iBin+2*i]   * binData[nx*iBin+2*i+1];
         }
         cc = cc/nBin;
         ss = ss/nBin;
         cs = cs/nBin;
         c0 = aveVis[2*i];
         s0 = aveVis[2*i+1];
         a0 = c0*c0 + s0*s0;
         errAmp[i] = sqrt((c0*c0*cc + s0*s0*ss + 2*c0*s0*cs)/
            (a0*nBin));
         errPhase[i] = degra* sqrt((c0*c0*ss +s0*s0*cc - 2*c0*s0*cs)/
            ((float)nBin))/a0;
      }
   } else {
      for (i=0;i<nWave;i++) { /* could not computer errors */
         errAmp[i]   = 0.;
         errPhase[i] = 0.;
      }
   }
/*
      prepare output visibility table to receive THE ANSWER
*/
   vTable = make_vis_table_def(outFile, nWave, "1776-07-04", "VLTI",
      "MIDI/PRISM", &ierr);
   vRow   = make_vis_row(vTable);
/* 
      fill in nonData items
*/
   vRow->targetId = 1;
   vRow->time     = 0.;
   vRow->mjd      = 0.;
   vRow->intTime  = totalExposure;
   vRow->ucoord   = u;
   vRow->vcoord   = v;
   vRow->station[0] = 1;
   vRow->station[1] = 2;

/*
      fill in Data
*/
   for (i=0; i<nWave; i++) {
      vRow->visAmp[i]    = visAmp[i];
      vRow->visAmpErr[i] = errAmp[i];
      vRow->visPhi[i]    = visPhase[i];
      vRow->visPhiErr[i] = errPhase[i];
      vRow->flag[i]      = 0; /* unflagged i.e. good data */
   }
   vRow->targetId = 1;
/*
      write row
*/
   i = write_vis_table_row(outFile, vTable, vRow, 1, &ierr);

/*
      shut down and shut up
*/
   i = close_vis_table(outFile, vTable, &ierr);
   kill_vis_row(vRow);

   i = close_imaging_fdata_table(inFile, idata, &ierr);
   kill_imaging_fdata_row(inData);
   free(frequency);
   free(goodData);
   free(binData);
   free(wave);
   free(band);
   free(flagTimes);
   free(aveVis);
   free(visAmp);
   free(visPhase);
   free(errAmp);
   free(errPhase);
   return 0;
}
